@@ -1,12 +1,13 @@
-import datetime
 import os
 import time
+import datetime
 import threading
-
-import requests
 import typing
-import yaml
+import requests
 import json
+import yaml
+from collections import defaultdict
+
 from . import zed, zjson  # TBD use upstream
 
 default_lake_url = os.environ.get("ZED_LAKE", "http://localhost:9867")
@@ -18,25 +19,26 @@ class Sync(threading.Thread):
     SKIP = 2
 
     def __init__(self,
-                 sources: list,
+                 sources: typing.List[str],
                  dest: str,
                  in_flow: str = "",
                  out_flow: str = "",
                  *,
                  poll_interval: float = -1,  # sec, <0: use push
                  eoio: bool = True,  # exactly-once in-order
+                 patch_source: bool = False,
                  owner: str = "sync",  # commit author
                  lake_url: str = default_lake_url,
                  client: zed.Client = None,
                  ):
         assert len(sources) > 0 and dest != ""
-
         self.sources = self._normalize(sources)
         self.dest = self._normalize_one(dest)
         self.in_flow = in_flow  # TBD multi-in_flow
         self.out_flow = out_flow
         self.poll_interval = poll_interval
         self.owner = owner
+        self.patch_source = patch_source
         # if eoio enabled, the sync agent will
         # process only those records that contain
         # a 'ts' field; XXX assume pool key is ts
@@ -63,6 +65,11 @@ class Sync(threading.Thread):
         self._stop_flag.set()
 
     def once(self):
+        records = self.read()
+        if len(records) != 0:
+            self.load(records)
+
+    def read(self) -> list:
         records = list()
         if self.eoio:
             self.query_str = self._make_query()
@@ -80,15 +87,17 @@ class Sync(threading.Thread):
                     self.source_ts[source] = max(max_ts, self.source_ts[source])
             else:
                 self.source_ts[source] = datetime.datetime.min
+        return records
+
+    def load(self, records: list):
         records = "\n".join(zjson.encode(records))
-        if len(records) > 0:
-            dest_pool, dest_branch = self._denormalize_one(self.dest)
-            self.client.load(
-                dest_pool, records,
-                branch_name=dest_branch,
-                commit_author=self.owner,
-                meta=self._source_ts_json(),
-            )
+        dest_pool, dest_branch = self._denormalize_one(self.dest)
+        self.client.load(
+            dest_pool, records,
+            branch_name=dest_branch,
+            commit_author=self.owner,
+            meta=self._source_ts_json(),
+        )
 
     def _event_loop(self):
         s = requests.Session()
@@ -117,13 +126,15 @@ class Sync(threading.Thread):
             cur_ts = self.source_ts.get(source, datetime.datetime.min)
             filter_flow = f"ts > {zjson.encode_datetime(cur_ts)} |" \
                 if self.eoio else ""
-            in_str += f"pool {source} => {filter_flow} fork (" \
+            patch_source_flow = f"put from := '{source}' |" \
+                if self.patch_source else ""
+            in_str += f"pool {source} => {filter_flow} {patch_source_flow} fork (" \
                       f"=> select max(ts) as max_ts | put __from := '{source}' " \
                       f"=> {'pass' if self.in_flow == '' else self.in_flow})"
             if len(self.sources) > 1:
                 in_str += "\n"
         in_str += ")\n"  # wrap up from clause
-        out_str = f"fork (=> has(__from) => " \
+        out_str = f"switch (case has(__from) => pass default => " \
                   f"{'pass' if self.out_flow == '' else self.out_flow})"
         return f"{in_str} | sort this | {out_str}"
 
@@ -193,6 +204,26 @@ class Sync(threading.Thread):
         return pool, branch
 
 
+class Watch(Sync):
+    """A destination-less sync that runs a UDF in once()."""
+
+    def __init__(self, fn: typing.Callable,
+                 source_ts=None, *args, **kwargs):
+        self.fn = fn
+        self.source_ts = source_ts
+
+        super().__init__(dest="none", *args, **kwargs)
+
+    def once(self):
+        self.fn(self.read())
+
+    def _fetch_source_ts(self) -> dict:
+        if self.source_ts is None:
+            return defaultdict(lambda: datetime.datetime.min)
+        else:
+            return self.source_ts
+
+
 def from_config(path: str) -> Sync:
     with open(path) as f:
         config = yaml.load(f, Loader=yaml.FullLoader)
