@@ -71,7 +71,7 @@ def run():
 
         # skip the last self-write
         # TBD for parallel reconciliation may need to lock rc.gen before patch
-        if gen == rc.skip_gen:
+        if gen == rc.skip_gen and rc.should_skip():
             digi.logger.info(f"skipped gen {gen} due to self-write")
             return
 
@@ -120,3 +120,4 @@ def run():
 
     # TBD graceful termination
     _ready, _stop = util.run_operator(_registry, log_level=digi.log_level)
+    digi.rc.start_data_watch()
