@@ -1,8 +1,9 @@
+import typing
 import inspect
 from collections import OrderedDict
 
+import digi
 import digi.util as util
-import digi.filter as filter_
 from digi.reconcile import rc
 
 """Filters."""
@@ -11,14 +12,16 @@ from digi.reconcile import rc
 def meta(*args, **kwargs):
     # if decorator not parameterized
     if len(args) >= 1 and callable(args[0]):
-        _attr(path="meta", *args, **kwargs)
+        register(path="meta", *args, **kwargs)
         return args[0]
 
     def decorator(fn):
         if len(args) >= 1:
-            _attr(fn, path="meta." + args[0], *args[1:], **kwargs)
+            register(fn, path="meta." + args[0], *args[1:], **kwargs)
         elif "path" in kwargs:
-            _attr(fn, path="meta." + kwargs.pop("path"), *args, **kwargs)
+            register(fn, path="meta." + kwargs.pop("path"), *args, **kwargs)
+        else:
+            register(fn, path="meta", *args, **kwargs)
         return fn
 
     return decorator
@@ -26,14 +29,16 @@ def meta(*args, **kwargs):
 
 def control(*args, **kwargs):
     if len(args) >= 1 and callable(args[0]):
-        _attr(path="control", *args, **kwargs)
+        register(path="control", *args, **kwargs)
         return args[0]
 
     def decorator(fn):
         if len(args) >= 1:
-            _attr(fn, path="control." + args[0], *args[1:], **kwargs)
+            register(fn, path="control." + args[0], *args[1:], **kwargs)
         elif "path" in kwargs:
-            _attr(fn, path="control." + kwargs.pop("path"), *args, **kwargs)
+            register(fn, path="control." + kwargs.pop("path"), *args, **kwargs)
+        else:
+            register(fn, path="control", *args, **kwargs)
         return fn
 
     return decorator
@@ -41,14 +46,16 @@ def control(*args, **kwargs):
 
 def data(*args, **kwargs):
     if len(args) >= 1 and callable(args[0]):
-        _attr(path="data", *args, **kwargs)
+        register(path="data", *args, **kwargs)
         return args[0]
 
     def decorator(fn):
         if len(args) >= 1:
-            _attr(fn, path="data." + args[0], *args[1:], **kwargs)
+            register(fn, path="data." + args[0], *args[1:], **kwargs)
         elif "path" in kwargs:
-            _attr(fn, path="data." + kwargs.pop("path"), *args, **kwargs)
+            register(fn, path="data." + kwargs.pop("path"), *args, **kwargs)
+        else:
+            register(fn, path="data", *args, **kwargs)
         return fn
 
     return decorator
@@ -56,29 +63,37 @@ def data(*args, **kwargs):
 
 def obs(*args, **kwargs):
     if len(args) >= 1 and callable(args[0]):
-        _attr(path="obs", *args, **kwargs)
+        register(path="obs", *args, **kwargs)
         return args[0]
 
     def decorator(fn):
         if len(args) >= 1:
-            _attr(fn, path="obs." + args[0], *args[1:], **kwargs)
+            register(fn, path="obs." + args[0], *args[1:], **kwargs)
         elif "path" in kwargs:
-            _attr(fn, path="obs." + kwargs.pop("path"), *args, **kwargs)
+            register(fn, path="obs." + kwargs.pop("path"), *args, **kwargs)
+        else:
+            register(fn, path="obs", *args, **kwargs)
         return fn
 
     return decorator
 
 
+# TODO fix dbox compatibility
+status = obs
+
+
 def mount(*args, **kwargs):
     if len(args) >= 1 and callable(args[0]):
-        _attr(path="mount", *args, **kwargs)
+        register(path="mount", *args, **kwargs)
         return args[0]
 
     def decorator(fn):
         if len(args) >= 1:
-            _attr(fn, path="mount." + args[0], *args[1:], **kwargs)
+            register(fn, path="mount." + args[0], *args[1:], **kwargs)
         elif "path" in kwargs:
-            _attr(fn, path="mount." + kwargs.pop("path"), *args, **kwargs)
+            register(fn, path="mount." + kwargs.pop("path"), *args, **kwargs)
+        else:
+            register(fn, path="mount", *args, **kwargs)
         return fn
 
     return decorator
@@ -86,14 +101,16 @@ def mount(*args, **kwargs):
 
 def ingress(*args, **kwargs):
     if len(args) >= 1 and callable(args[0]):
-        _attr(path="ingress", *args, **kwargs)
+        register(path="ingress", *args, **kwargs)
         return args[0]
 
     def decorator(fn):
         if len(args) >= 1:
-            _attr(fn, path="ingress." + args[0], *args[1:], **kwargs)
+            register(fn, path="ingress." + args[0], *args[1:], **kwargs)
         elif "path" in kwargs:
-            _attr(fn, path="ingress." + kwargs.pop("path"), *args, **kwargs)
+            register(fn, path="ingress." + kwargs.pop("path"), *args, **kwargs)
+        else:
+            register(fn, path="ingress", *args, **kwargs)
         return fn
 
     return decorator
@@ -101,36 +118,34 @@ def ingress(*args, **kwargs):
 
 def egress(*args, **kwargs):
     if len(args) >= 1 and callable(args[0]):
-        _attr(path="egress", *args, **kwargs)
+        register(path="egress", *args, **kwargs)
         return args[0]
 
     def decorator(fn):
         if len(args) >= 1:
-            _attr(fn, path="egress." + args[0], *args[1:], **kwargs)
+            register(fn, path="egress." + args[0], *args[1:], **kwargs)
         elif "path" in kwargs:
-            _attr(fn, path="egress." + kwargs.pop("path"), *args, **kwargs)
+            register(fn, path="egress." + kwargs.pop("path"), *args, **kwargs)
+        else:
+            register(fn, path="egress", *args, **kwargs)
         return fn
 
     return decorator
 
 
-# XXX test path
-def attr(*args, **kwargs):
+def model(*args, **kwargs):
     if len(args) >= 1 and callable(args[0]):
-        _attr(path=".", *args, **kwargs)
+        register(path=".", *args, **kwargs)
         return args[0]
 
     def decorator(fn):
-        _attr(fn, *args, **kwargs)
+        register(fn, *args, **kwargs)
         return fn
 
     return decorator
 
 
-model = attr
-
-
-def _attr(fn, path=".", prio=0):
+def register(fn, path=".", prio=0, cond=digi.filter.changed):
     # preprocess the path str -> tuple of str
     _path = list()
     ps = path.split(".")
@@ -215,6 +230,11 @@ def _attr(fn, path=".", prio=0):
             kwarg_filter.update({"diff": p})
             args[p] = None
 
+    for p in ["meta"]:
+        if p in sig.parameters:
+            kwarg_filter.update({"meta": p})
+            args[p] = None
+
     for p in ["typ", "child_typ"]:
         if p in sig.parameters:
             kwarg_filter.update({"typ": p})
@@ -239,12 +259,14 @@ def _attr(fn, path=".", prio=0):
             kwarg_filter["back_prop"] = k
         elif i == 7:
             kwarg_filter["diff"] = k
+        elif i == 8:
+            kwarg_filter["meta"] = k
         else:
             break
 
     def wrapper_fn(subview, proc_view, view,
                    old_view, mount, obs, back_prop,
-                   diff):
+                   diff, meta):
         kwargs = dict()
         for _k, _v in [("subview", subview),
                        ("proc_view", proc_view),
@@ -254,6 +276,7 @@ def _attr(fn, path=".", prio=0):
                        ("obs", obs),
                        ("back_prop", back_prop),
                        ("diff", diff),
+                       ("meta", meta),
                        ("typ", child_typ),
                        ]:
             if _k in kwarg_filter:
@@ -262,9 +285,10 @@ def _attr(fn, path=".", prio=0):
 
     rc.add(handler=wrapper_fn,
            priority=prio,
-           condition=filter_.has_diff,
+           condition=cond,
            path=_path)
 
+
 def mount_change(diff, gvr=None) -> bool:
     """Detect whether the diffs contain newly mounted digi."""
     for _diff in diff:
@@ -279,3 +303,47 @@ def mount_change(diff, gvr=None) -> bool:
         elif path[:3] == ("spec", "mount", gvr):
             return True
     return False
+
+
+def pool(*args, **kwargs):
+    # TBD load max_ts on restart; the max_ts should be stored
+    #   at a persistent location and written atomically, e.g.,
+    #   the digi.model or digi.pool
+
+    class DelayedWatch:
+        def __init__(self, fn, *args, **kwargs):
+            self.fn = fn
+            self.args = args
+            self.kwargs = kwargs
+            self.watch = None
+
+        def start(self):
+            self.watch = digi.pool.watch(
+                self.fn, *self.args, **self.kwargs
+            )
+            self.watch.start()
+
+        def stop(self):
+            self.watch.stop()
+
+    # @pool
+    if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
+        digi.rc.add_data_watch(
+            name=watch_name(args[0]),
+            watch=DelayedWatch(args[0], *args, **kwargs),
+        )
+        return args[0]
+
+    # @pool(*args, **kwargs)
+    def decorator(fn):
+        digi.rc.add_data_watch(
+            name=watch_name(fn),
+            watch=DelayedWatch(fn, *args, **kwargs),
+        )
+        return fn
+
+    return decorator
+
+
+def watch_name(fn):
+    return fn.__name__.lstrip("gen_")
